r"""Utilities
"""
def getMD5ForIterable(Iterable):
	import hashlib

	hashMD5 = hashlib.md5()

	for chunk in Iterable:
		hashMD5.update(chunk)

	return hashMD5.hexdigest()

def getMD5(filePath): # #From: http://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file
	r"""Returns the MD5 digest of the given file.
	"""
	from laufire.filesys import iterateContent

	return getMD5ForIterable(iterateContent(filePath))

def getTimeString(useLocal=False):
	r"""Returns the current time as a path friendly string.
	"""
	from datetime import datetime

	return (datetime.now if useLocal else datetime.utcnow)().strftime('%y%m%d%H%M%S%f')[:-3]

def getRandomString(length=8):
	import random
	import string

	return ''.join(random.choice(string.ascii_uppercase) for _ in range(length))

def getFreeName(iterable):
	r"""Gets a free name for the given iterable.
	"""
	while True:
		name = getRandomString()

		if name not in iterable:
			return name



r"""
FileSys
=======

	A module to help with files system operations.

Notes
-----

	* Path removals generally require an ancestor to be specified, so to avoid accidental deletes.
	* Many calls remove the target by default; these might not work outside the fsRoot; use removePath with requiredAncestor to achieve the same effect.
	* The module shutill isn't used as, some of its operations could be dangerous. Ex: shutil.rmtree will remove the files with-in dir junctions instead of merely removing the junction.
	* To avoid conflicts while writing outside the fsRoot, remove the target first.
	* Soft links (symlinks) are the default as they are more robust (they work even after the source is replaced).

Caution
-------

	* Debugging absolute paths could be a security risk. Hence set the appropriate logLevel.

Pending
-------

	* Check: Could links be removed, even outside the fsRoot?
	* The module doesn't handle unicode file-names. So does the package zipfile of Python2. It depends upon the type of the input path strings. ie: To call collectPaths on a dir with an unicode file name, the source dir should also be in unicode.
	* Support file encodings.
	* Add a function to copy file attributes, etc.
	* Add an isOk call to verify the correctness of the files (to avoid unresolved links etc).
	* Check: Use absPaths for robustness of the calls.
"""

import os
from os import mkdir, makedirs, unlink, rmdir
from os.path import abspath, basename, commonprefix, dirname, exists, isdir, isfile, join as pathJoin, normpath, split as pathSplit, splitext
import re

from laufire.flow import forgive
from laufire.logger import debug
from laufire.utils import getRandomString, getTimeString
from laufire.helpers.filesys import link, symlink, rmlink, isLinkedDir

# State
fsRoot = '.' # Risky filesystem operations such as removePath are limited to fsRoot.
sep = os.sep

# Data
Ext2Opener = {'zip': ('zipfile', 'ZipFile'), 'gz': 'gzip'} # #Pending: Instead of having module, object pairs import objects (for that write a support module. ie: import_obj('zipfile.ZipFile')

# Helpers
def globToRe(pattern):
	ret = r'^'
	pattern += '|'

	i = 0

	while i < len(pattern) - 1:
		c = pattern[i]

		if c != '*':
			ret += re.escape(c)

		else:
			if pattern[i + 1] != '*':
				ret += r'[^\/]*'

			else:
				ret += r'.*'
				i += 1

		i += 1

	ret += r'$'

	return ret

makeMissingDir = lambda path: exists(path) or makeDir(path)

def rmtree(tgtPath):
	r"""Removes a dir tree. It can unlink junctions (even as sub-dirs), without removing the descendant files.
	"""
	absPath = abspath(tgtPath)

	for root, Dirs, Files in os.walk(absPath):
		for file in Files:
			unlink(pathJoin(root, file))

		for dir in Dirs:
			dir = pathJoin(root, dir)

			if isLinkedDir(dir):
				rmlink(dir)

			else:
				rmtree(dir)

	rmdir(tgtPath)

def _makeLink(srcPath, tgtPath, pathType, hardLink):
	debug('link: %s => %s' % (srcPath, tgtPath))

	(link if hardLink and pathType == 1 else symlink)(abspath(srcPath), tgtPath) # #Note: Dirs can't be hard-linked.

def _removePath(tgtPath): # #Pending: Check if the call could be made more efficient.
	if not exists(tgtPath):
		return 1

	if isfile(tgtPath):
		unlink(tgtPath)

	elif isLinkedDir(tgtPath):
		rmlink(tgtPath)

	elif isdir(tgtPath):
		rmtree(tgtPath)

	else:
		return 1 # error

def _getOpener(ext):
	opener = Ext2Opener.get(ext)

	if opener:
		from importlib import import_module

		# #Pending: In the case of opening a ZipFile, if a filename within the archive isn't provided, use the first file.

		if hasattr(opener, 'upper'): # Only a module name is available. #Pending: Implement proper string validation.
			return import_module(opener).open

		return lambda filePath: getattr(import_module(opener[0]), opener[1])(filePath).open

	else:
		return lambda filePath: open(filePath, 'rb')

def doNoting(firstArg, *dummy, **dummy1):
	return firstArg

# Exports
## Path functions
stdPath = (lambda path: path.replace('\\', '/')) if sep != '/' else doNoting # Standardizes the given path.

def pair(src, tgt, postFix):
	r"""Returns a pair of the given post-fix affixed source and target paths.
	"""
	return joinPaths(src, postFix), joinPaths(tgt, postFix)

def joinPaths(*Paths):
	r"""
	Joins the given Paths.
	"""
	ret = '/'.join(Paths)

	return ret if Paths[0] else ret[1:]

def resolve(basePath, relation):
	r"""Resolves a relative path of the given basePath.

	Args:
		basePath (path): The path of the base.
		relation (str): The relation. Ex: '..', '../ops' etc.
	"""
	return abspath(pathJoin(basePath, relation))

def getFreeFilePath(parentDir, length=8):
	if exists(parentDir):
		while True:
			freePath = joinPaths(parentDir, getRandomString(length))

			if not exists(freePath):
				return freePath

def getPathType(path):
	r"""#Note: The path-types are:
		0: missing
		1: file
		2: dir
		3: linked dir
	"""
	ret = 3

	for func in [isLinkedDir, isdir, isfile]:
		if func(path):
			break

		ret -= 1

	return ret

def isLocked(filePath, tempPath=None): # #Pending: Check, whether there is a proper and robust way to check the lock status, instead of renaming the path.
	r"""Checks whether the given path is locked.
	"""
	if not exists(filePath):
		return

	if not tempPath:
		while True:
			tempPath = '%s.%s' % (filePath, getRandomString())

			if not exists(tempPath):
				break

	if forgive(lambda: rename(filePath, tempPath)):
		return True

	rename(tempPath, filePath)

	return False

def isContainer(path):
	return isdir(path) or isLinkedDir(path)

def isDescendant(probableDescendant, requiredAncestor):
	r"""Checks whether the given path is a descendant of another.

	Args:
		probableDescendant (str): The absolute path of the probable descendant.
		requiredAncestor (str): The absolute path of the required ancestor.
	"""
	requiredAncestor = abspath(requiredAncestor)

	return commonprefix([abspath(probableDescendant), requiredAncestor]) == requiredAncestor

def requireAncestor(path, requiredAncestor=None):
	r"""Ensures that the given path is a decendant of the required ancestor.
	"""
	ancestor = requiredAncestor or fsRoot

	if not isDescendant(path, ancestor):
		raise Exception('"%s" is not a descendant of "%s"' % (path, ancestor))

def getAncestor(path, ancestorName):
	r"""
	Returns the first ancestor of the given path, which has the given name.
	"""
	Lineage = abspath(path).split(sep)
	l = len(Lineage)

	while l:
		l -= 1

		if Lineage[l] != ancestorName:
			continue

		return joinPaths(*Lineage[:l + 1])

def collectPaths(base='.', pattern='**', regex=False, followlinks=True):
	r"""Returns two lists containing dirs and files of the given base dir.

	Args:
		base (str): The path to scan for.
		pattern:
			(glob): A '|' separated list of globs. Includes and excludes are separated by a '!'. Defaults to all ('**').
			(regex): A '|' separated list of regex. Includes and excludes are separated by a '$$$'. Defaults to all ('.*').
		regex (bool, False): When set to True, Includes and Excludes are parsed as regular expressions, instead of as globs.

	Glob guide:
		** : anything  (including empty strings)
		*  : anything but a slash (including empty strings)
		other characters : as is

	#From: http://stackoverflow.com/questions/5141437/filtering-os-walk-dirs-and-files
	#Note: The globs are not regular globs. But a simplified versions.
	#Note: Exclusions override inclusions.
	"""
	if regex:
		Split = pattern.split('$$$')
		includes = Split[0] or r'.*'
		excludes = Split[1] if len(Split) > 1 else None

	else:
		# transform globs to regular expressions
		Split = pattern.split('!')
		includes = r'|'.join([globToRe(x) for x in Split[0].split('|')]) if Split else r'.*'
		excludes = r'|'.join([globToRe(x) for x in Split[1].split('|')]) if len(Split) > 1 else None

	baseLen = len(base)
	match = re.match

	for root, Dirs, Files in os.walk(base, followlinks=followlinks):

		prefix = stdPath(root[baseLen+1:])

		Joined = {d: joinPaths(prefix, d) for d in Dirs}

		if excludes:
			Dirs[:] = [d for d, j in Joined.iteritems() if not match(excludes, j)] # Exclude dirs from recursion.

		for d, j in [(d1, j1) for d1, j1 in Joined.iteritems() if d1 in Dirs and match(includes, j1)]: # Yield resulting dirs.
			yield j, 2 if isdir(joinPaths(root, d)) else 3 # Check whether the path is a dir or a link.

		Files = [joinPaths(prefix, f) for f in Files]

		for file in [f for f in Files if not (excludes and match(excludes, f)) and match(includes, f)]:
			yield file, 1

def glob(base='.', pattern='**', pathType=None):
	r"""Yields the paths under the given dir matching the given glob pattern.
	"""
	jp = joinPaths

	if pathType is None:
		for path, dummy in collectPaths(base, pattern):
			yield jp(base, path)

	else:
		for path, _pathType in collectPaths(base, pattern):
			if _pathType == pathType:
				yield jp(base, path)

def getPathPairs(source, target, pattern='**', regex=False):
	r"""Iterates over the given patterns and yields a tuple with source and destination paths.
	"""
	for path, dummy in collectPaths(source, pattern, regex):
		yield pair(source, target, path)

## File manipulation functions
def removePath(tgtPath, requiredAncestor=None, forced=False):
	r"""Removes any given file / dir / junction.

	Args:
		tgtPath (path): The path to remove.
		requiredAncestor (path): The target will only be removed if it's a descendant of this dir. Defaults to the global attr fsRoot.
		forced (bool): When set to true, an ancestor won't be required.
	"""
	if not forced:
		requireAncestor(tgtPath, requiredAncestor)

	debug('remove: %s' % tgtPath)
	return _removePath(tgtPath)

def makeDir(path):
	debug('making: %s' % path)
	mkdir(path)

def makeLink(srcPath, tgtPath, autoClean=True, hardLink=False):
	r"""Links the given paths.
	"""
	if autoClean:
		removePath(tgtPath)

	ensureParent(tgtPath)

	_makeLink(srcPath, tgtPath, getPathType(srcPath), hardLink)

def rename(srcPath, tgtPath, autoClean=True):
	r"""Renames any given file / dir / junction.
	"""
	if autoClean:
		removePath(tgtPath)

	debug('rename: %s => %s' % (srcPath, tgtPath))

	os.rename(srcPath, tgtPath)

def copy(srcPath, tgtPath, pattern='**', regex=False, autoClean=True):
	r"""Copies one path to another.
	"""
	debug('copy: %s => %s' % (srcPath, tgtPath))

	if autoClean:
		removePath(tgtPath) # #Note: This also ensures that the target is under fsRoot.

	pathType = getPathType(srcPath)

	if not pathType:
		ensureParent(tgtPath)

	if pathType == 1:
		copyContent(srcPath, tgtPath)

	else:
		dirMaker = makeDir if autoClean else makeMissingDir
		copier = lambda src, tgt: copyContent(src, tgt, autoClean=False) if autoClean or not exists(tgtPath) else copyContent # File safety isn't a concern inside a missing dir.

		makeMissingDir(tgtPath)

		for Path in collectPaths(srcPath, pattern, regex):
			if Path[1] != 1:
				dirMaker(joinPaths(tgtPath, Path[0]))

			else:
				copier(*pair(srcPath, tgtPath, Path[0]))

def linkTree(srcPath, tgtPath, pattern='**', regex=False, autoClean=True, hardLink=False):
	r"""Re-creates the structure of the source at the target by creating dirs and linking files.
	"""
	_srcPath = abspath(srcPath) # Link sources should be abs-paths.
	dirMaker = makeDir if autoClean else makeMissingDir
	linkWorker = link if hardLink else symlink # Hard links aren't the default, as they can't work across drives.
	linker = linkWorker if autoClean or not exists(tgtPath) else lambda srcPath, tgtPath: (removePath(tgtPath), linkWorker(srcPath, tgtPath)) # File safety isn't a concern inside a missing dir.
	parentMaker = ensureParent if pattern != '**' and not regex else doNoting

	(ensureCleanDir if autoClean else ensureDir)(tgtPath)

	for Path in collectPaths(srcPath, pattern, regex):
		if Path[1] != 1:
			dirMaker(joinPaths(tgtPath, Path[0]))

		else:
			path = Path[0]
			_tgtPath = joinPaths(tgtPath, path)
			debug('link: %s => %s' % (joinPaths(srcPath, path), _tgtPath))
			parentMaker(_tgtPath)
			linker(joinPaths(_srcPath, path), _tgtPath)

def ensureParent(childPath):
	r"""Ensures the parent dir of the given childPathexists.

		This function is provided to ease the use of other file copying tasks.

		#Note: makedirs isn't used, as it doesn't report properly on occupied names.
	"""
	parentPath = dirname(normpath(childPath))
	Paths = []

	while parentPath and not isContainer(parentPath):
		assert not exists(parentPath), 'Parent path is occupied: %s'% parentPath # The path is occupied by a file / link.
		Paths.insert(0, parentPath)
		parentPath = dirname(parentPath)

	for path in Paths:
		makeDir(path)

def ensureDir(dir):
	r"""Ensures that the given dir is available.
	"""
	if not exists(dir):
		ensureParent(dir)
		makeDir(dir)

def ensureCleanDir(dir, requiredAncestor=None):
	r"""Ensures that the given dir is clean and available.
	"""
	removePath(dir, requiredAncestor)
	makedirs(dir)

## Content Functions
def getLines(filePath, start=-1, end=-1, ext=None, Args=None):
	r"""Yields lines from various file formats like zip, gzip etc.
	"""
	if not ext:
		ext = splitext(filePath)[1]

	n = 0

	if end < 0:
		end = float('inf')

	else:
		end += 2

	opener = _getOpener(ext.lower()[1:])(filePath)

	for line in opener if not Args else opener(*Args):
		n += 1

		if n > start:
			if n < end:
				yield line

			else:
				break

def getContent(filePath):
	r"""Reads a file and returns its content.
	"""
	return open(filePath, 'rb').read()

def iterateContent(filePath, width=4096):
	r"""Reads the given file as small chunks. This could be used to read large files without buffer overruns.
	"""
	with open(filePath, 'rb') as file:
		for chunk in iter(lambda: file.read(width), b''):
			yield chunk

def setContent(filePath, content, autoClean=True):
	r"""Fills the given file with the given content.
	"""
	if (removePath(filePath) == 1 if autoClean else True): # Ensure parent only if the path is not removed.
		ensureParent(filePath)

	open(filePath, 'wb').write(content)

def appendContent(filePath, content):
	r"""Appends the given file with the given content.
	"""
	if not exists(filePath):
		return setContent(filePath, content)

	open(filePath, 'ab').write(content)

def copyContent(srcPath, tgtPath, autoClean=True):
	r"""
	Copies the content of one file to another.

	# #Note: Unlike shutil.copy attributes aren't copied.
	"""
	debug('copy: %s => %s' % (srcPath, tgtPath))

	if (removePath(tgtPath) == 1 if autoClean else True): # Ensure parent only if the path is not removed.
		ensureParent(tgtPath)

	with open(tgtPath, 'wb') as tgt:
		for chunk in iterateContent(srcPath):
			tgt.write(chunk)

def compress(srcPath, tgtPath): # #Note: shutil.make_archive isn't used, due to its forcing of the zip extension and due to the need for maintaing a compression standard.
	if not exists(srcPath):
		raise('No such path: %s' % srcPath)

	ensureParent(tgtPath)

	from zipfile import ZipFile, ZIP_DEFLATED

	ZipFileObj = ZipFile(tgtPath, 'w', ZIP_DEFLATED)

	cwd = os.getcwd() # The CWD circus is to reduce the relpath calls.

	if isContainer(srcPath):
		os.chdir(srcPath)

		for root, dummy1, Files in os.walk('.', followlinks=True):
			for file in Files:
				ZipFileObj.write(pathJoin(root, file))

	else:
		dir, name = pathSplit(srcPath)
		os.chdir(dir)

		ZipFileObj.write(name)

	os.chdir(cwd)

	ZipFileObj.close()

def extract(srcPath, tgtPath): # #Note: The tagertPath points to the extraction root. Hence, it should be a dir.
	from zipfile import ZipFile

	with ZipFile(srcPath, 'r') as Z:
		Z.extractall(tgtPath)

def backup(srcPath, backupBase=None, addTimeString=True, keepOriginal=False): # #Pending: Change the name of the call to preserve or safeguard. As the current name could be misleading.
	r"""Backs up the given path to a temporary location, so that the returned path could be later used to restore it to the original location.

	When backupBase isn't given the path is renamed, not moved.
	# #Note: This call along with restore, could preserve links.
	"""
	if not backupBase:
		backupBase = dirname(abspath(srcPath))

	tgtPath = '%s%s.bak' % (pathJoin(backupBase, basename(srcPath)), ('.%s' % getTimeString()) if addTimeString else '')
	removePath(tgtPath, backupBase)

	(copy if keepOriginal else os.rename)(srcPath, tgtPath)

	return tgtPath

def restore(backupPath, srcPath=None): #Pending: Add a way to handle time strings. Better yet a make calss for backup (may be as a separate module).
	if not srcPath:
		srcPath = splitext(backupPath)[0]

	_removePath(srcPath)
	os.rename(backupPath, srcPath)

# Init
def setup(Project):
	global fsRoot

	_temp = Project.fsRoot

	fsRoot = abspath(_temp)
	debug('fsRoot: %s' % _temp)

from laufire.initializer import loadProjectSettings #pylint: disable=C0413
loadProjectSettings(setup)



r"""A helper for SSH operations.

#Note: This module needs the SSHConfig to be in Project.Config['SSH']
#Later: Ad support for recursive downloads.
"""
import json
import errno

from os import listdir
from os.path import basename, split as pathSplit

import paramiko
from laufire.extensions import Lazy
from laufire.filesys import getPathType, joinPaths, pair
from laufire.flow import forgive, retry
from laufire.logger import debug
from laufire.shell import assertShell

# Helpers
def _upload(SFTP, localPath, remotePath):

	debug('uploading %s to %s' % (localPath, remotePath))

	pathType = getPathType(localPath)

	if pathType == 1: # file
		retry(lambda: SFTP.put(localPath, remotePath) or 1) # #Note: 1 is returned to notify retry of a success.

	elif pathType > 1: # dir / junction
		err = forgive(lambda: mkdirs(SFTP, remotePath))

		if err and not isinstance(err, IOError):
			raise err

		for item in listdir(localPath): # #Note: With dir uploads, the uploads are merged (added / overwritten) with existing paths.
			retry(lambda: _upload(SFTP, *pair(localPath, remotePath, item))) #pylint: disable=cell-var-from-loop

	else: # Path doesn't exist.
		raise Exception('Invalid source path: %s' % localPath)

	return remotePath

def mkdirs(SFTP, remotePath):
	currentPath = remotePath

	Skipped = []

	while True:
		err = forgive(lambda: SFTP.mkdir(currentPath, 511)) # #Note: Existense isn't checked for to reduce the number of remote calls.

		if err:
			if err.errno is None: # The dir exists.
				return

			elif not isinstance(err, IOError): # #Pending: Check: Permission errors could result in infinite loops.
				raise err

			else:
				# Try to create the parent path.
				currentPath, skipped = pathSplit(currentPath)
				Skipped.append(skipped)

				if not currentPath or currentPath == '/':
					raise Exception('Failed to create the dir: %s' % remotePath)

		else:
			if not Skipped:
				break

			else:
				currentPath += '/%s' % Skipped.pop(0)

def getTgtName(tgtName, srcPath):
	return tgtName if tgtName else basename(srcPath)

# Classes
class SSHClient(paramiko.SSHClient):
	r"""An abstraction layer over the SSH client.
	"""
	def __init__(self, SSHConfig):
		paramiko.SSHClient.__init__(self)

		self.set_missing_host_key_policy(paramiko.AutoAddPolicy())
		self.load_system_host_keys()
		self.connect(SSHConfig['host'], username=SSHConfig['username'], password=SSHConfig['password'])
		self._SFTP = self.open_sftp()

	def __del__(self):
		self._SFTP.close()

	def __getattr__(self, attr):
		r"""
		Allows to access the methods of the underlying SFTP connection.
		"""
		return getattr(self._SFTP, attr)

	def download(self, remotePath, localPath=''):
		debug('downloading %s to %s' % (remotePath, localPath))
		SFTP = self.open_sftp()
		localPath = getTgtName(localPath, remotePath)
		SFTP.get(remotePath, localPath)
		SFTP.close()

		return localPath

	def upload(self, localPath, remotePath):
		remotePath = getTgtName(remotePath, localPath)

		SFTP = self.open_sftp()
		_upload(SFTP, localPath, remotePath)
		SFTP.close()

		return remotePath

	def exists(self, path):
		err = forgive(lambda: self._SFTP.stat(path))

		return True if not err else (isinstance(err, IOError) and err.errno == errno.ENOENT)

	def execute(self, command):
		debug(command)
		dummy, stdout, stderr = self.exec_command(command)

		return {

			'code': stdout.channel.recv_exit_status(),
			'out': stdout.read(),
			'err': stderr.read(),
		}

class SSHBridge:
	r"""Bridges with the SSH gateway of the remote host.
	"""
	def __init__(self, Config):
		self.Config = Config
		self.Client = Lazy(SSHClient, Config['SSH']) # #Note: SSHBridge is initialized as lazy class, as parmiko cannot connect to the server when modules are being loaded, due to some internals of threading.

	def __getattr__(self, attr):
		r"""
		Allows access to the methods of the underlying client.
		"""
		return getattr(self.Client, attr)

	def iexecute(self, command):
		r"""Interpolates the command with the Config, before executing.
		"""
		return self.Client.execute(command.format(**self.Config['Gateway']))

	def callScript(self, ecCommand):
		out = assertShell(self.iexecute('{Python[binary]} {Paths[private]}/%s' % ecCommand))
		return json.loads(out) if out else None

	def upload(self, srcPath, tgtPath=''):
		r"""Uploads through the gateway are done to the temp dir by default.
		"""
		if not tgtPath:
			tgtPath = joinPaths(self.Config['Gateway']['Paths']['temp'], getTgtName(None, srcPath))

		else:
			tgtPath = tgtPath.format(**self.Config['Gateway'])

		return self.Client.upload(srcPath, tgtPath)



r"""
A simple tool to help with re-running failed tasks.

#Note: This module could be called from the command line.
#Usage: python -m laufire.retry "command" <cwd> <repeats> <delay>
"""
# Exports
def retryCommand(command, repeat, delay, **KWArgs):
	r"""
	The keyword arguments are for shell.run.
	"""
	from laufire.flow import retry
	from laufire.shell import run
	from time import time

	State = {'tries': 0, 'exitCode': None}

	def worker():
		exitCode = run(command, **KWArgs)

		State['tries'] += 1
		State['exitCode'] = exitCode

		if not exitCode:
			return 1

	start = time()

	retry(worker, repeat, delay)

	return State['exitCode'], State['tries'], time() - start

# Main
def main():
	import sys
	from laufire.logger import log

	Args = sys.argv[1:]
	command = Args[0]
	cwd = Args[1] if len(Args) > 1 else '.'
	repeat = int(Args[2]) if len(Args) > 2 else 1
	delay = int(Args[3]) if len(Args) > 3 else 0

	exitCode, tries, duration = retryCommand(command, repeat, delay, cwd=cwd)

	log('tries: %s' % tries, 'LIGHTGREEN_EX')
	log('duration: %s' % duration, 'LIGHTYELLOW_EX')

	sys.exit(exitCode)

if __name__ == '__main__':
	main()



r"""A module to help with storing validated data.

#Pending: Fix: ConfiguredStore doesn't seem to be cleared, along with their namespace. To simulate the condition, call the store from function and check for the locked state of the db, even after the function exits.
#Later: Fix: Calling var from the command line doesn't pretty print the branches.
#Later: Fix: Make dump to use pPrint.
#Later: The commad setup followed by any text triggers a overwrite, use ec inputs to conver the arg to boolean.
#Later: Rename the method Store.get into Store.gather or something else, to avoid confusing it with the get method of Dictionaries, as Stores also support keys, like dictionaries.
#Later: vars could use Classes to define separate hooks for get, set etc; instead of using the current, two param function hooks. Note, add this feature, while retaining the single function hook, without commands, which is to be called on init and set.
#Later: Think of using **with blocks** for declaration, instead of decorators, as the syntax looks cleaner. #Refer: http://stackoverflow.com/questions/1255914/finding-functions-defined-in-a-with-block
#Later: Branch level hooks that are to be called when any of its Children are changed.

"""

#pylint: disable=redefined-argument-from-local

import re

from os.path import exists

from json import loads, dumps
from collections import OrderedDict

from ec.utils import get
from laufire.sqlitex import SQLiteDB, SQLiteSimpleTable
from laufire.extensions import combine
from laufire.parser import parse as _parse

# State
State = []
keyPartPattern = re.compile(r'([^/]+/)')
Commands = ['setup', 'var', 'dump']

# Helpers
def getName(Obj, Dict):
	if not 'name' in Dict:
		Dict['name'] = Obj.__name__

	return Dict['name']

def getSTDIN():
	from sys import stdin

	if stdin.isatty():
		return

	b = b''

	for chunk in iter(lambda: stdin.read(4096), b''):
		b += chunk

	return b or None

# Workers
def collectChildren(Obj):
	AttrDict = {} # Used to get the name of vars, without hooks.

	for attr in [attr for attr in dir(Obj) if attr[0] != '_']:
		Child = getattr(Obj, attr)

		AttrDict[getattr(Child, 'im_func', Child)] = attr # #Note: Functions defined within the classes of the branches are altered as methods. Hence im_func is checked for, to find the original child.

	childCount = len([attr for attr in dir(Obj) if attr[0] != '_'])
	Children = []

	while childCount:
		childCount -= 1

		childName, Config, Child = State.pop()

		if childName is None:
			childName = AttrDict[Child]

		if not 'name' in Config:
			Config['name'] = childName

		Children.insert(0, (childName, Config,))

	return OrderedDict(Children)

def getRoute(branch, path):
	return ('%s/%s' % (branch, path)) if branch else path

def getConfigsFromDict(Dict, branch, Buffer):
	Configs = Buffer['Configs']
	Routes = []
	Configs[branch] = {'Routes': Routes}

	for key, value in Dict.iteritems():
		route = getRoute(branch, key)
		Routes.append(route)

		if  hasattr(value, 'iteritems'):
			getConfigsFromDict(value, route, Buffer)

		else:
			Buffer['Values'][route] = value

def processCollected(Collected, branch, Buffer):
	for key, Item in Collected.iteritems():
		Children = Item.get('Children', None)

		route = getRoute(branch, key)

		if Children is not None: # We've got a branch.
			Config = Item['Config']
			Buffer['Configs'][route] = Config
			Config['Routes'] = [getRoute(route, k) for k in Children]
			processCollected(Children, route, Buffer)

		elif 'Data' in Item: # We've got a data dictionary.
			getConfigsFromDict(Item['Data'], route, Buffer)

		elif 'value' in Item: # We've got a read-only value.
			Buffer['Values'][route] = Item['value']

		else: # We've got a var.
			Buffer['Configs'][route] = Item

def getStoreTable(Config):
	filePath = Config['filePath']
	tableName = Config.get('tableName', 'ecstore')

	DB = SQLiteDB(filePath)
	DB.execute("CREATE TABLE IF NOT EXISTS %s (`route` TEXT PRIMARY KEY, value TEXT)" % tableName)
	DB.close()

	return SQLiteSimpleTable(filePath, tableName, 'route')

def split(route):
	i = route.rfind('/')
	return route[:i] if i > 0 else '', route[i + 1:]

def getBranch(route):
	i = route.rfind('/')
	return route[:i] if i > 0 else ''

def getLeaf(route):
	return route[route.rfind('/') + 1:]

def processCommand(Store):
	import sys
	Argv = sys.argv[1:]

	if not Argv:
		command = 'setup'

	else:
		command = Argv.pop(0)

		if command not in Commands:
			raise Exception('Command "%s" is not recognized.' % command)

	ret = getattr(Store, command)(*Argv)

	if ret is not None:
		print ret

# Classes
class ReadOnlyStore:
	def __init__(self, **Config):
		r"""
		Reads the values from the given Store.

		Config:
			filePath: The path to the store.
			tableName (str): The table name of the store, defaults to ecstore.
		"""
		Store = SQLiteSimpleTable(Config['filePath'], Config.get('tableName', 'ecstore'), 'route')
		self._Values = Values = {k: loads(v) for k, v in Store.getCol('value').iteritems()}
		Store.close()

		Routes = Values.keys()

		# Add branch configs.
		while Routes:
			currentRoute = Routes.pop(0)
			branch, leaf = split(currentRoute)

			if not leaf: # The route points to the root, hence skip processing it to avoid infinite looping.
				continue

			Branch = Values.get(branch)

			if not Branch:
				Values[branch] = Branch = {'Routes': []}
				Routes.append(branch)

			Branch['Routes'].append(currentRoute)

	def __getitem__(self, route):
		return self.var(route)

	def var(self, route):
		Value = self._Values[route]

		if hasattr(Value, '__getitem__') and 'Routes' in Value: # Return the values from the Children
			Ret = {}

			for i in Value['Routes']:
				Ret[getLeaf(i)] = self.var(i)

			return Ret

		return self._Values[route]

class ConfiguredStore:
	def __init__(self, Buffer, Config):
		self._Configs = Configs = Buffer['Configs']
		self._Values = Values = Buffer['Values']
		self._Store = Store = getStoreTable(Config)

		StoreValues = Store.getCol('value')

		for route, value in Values.iteritems(): # Write any parsed values to the DB, so that the DB could be shared without the parsed source.
			self._set(route, value)

		for route, value in StoreValues.iteritems():
			if route in Configs:
				Values[route] = loads(value)

			elif route not in Values: # Delete residual routes from the DB.
				Store.delete(route)

		for key, value in Values.iteritems():
			Config = Configs.get(key)

			if Config and 'live' in Config:
				Config['hook'](value, 'init')

	def __del__(self):
		if hasattr(self, '_Store'):
			self.close()

	def var(self, route, value=None): #pylint: disable=W0221
		Config = self._Configs.get(route)

		if value is None:
			value = getSTDIN()

		if value is None:
			if not Config:
				return self._Values[route]

			if 'Routes' in Config: # Return the values from the Children.
				Ret = {}
				for i in Config['Routes']:
					Ret[getLeaf(i)] = self.var(i)

				return Ret

			else:
				storeValue = self._Values[route]

				if Config.get('live'): # Pass the value to the hook.
					ret = Config['hook'](storeValue, 'get')

					if ret is not None:
						return ret

				return storeValue

		else:

			if not Config:
				raise Exception('Cannot set the value of a read-only var.')

			elif 'Routes' in Config:
				raise Exception('The route poins to a branch, not a var.')

			if 'type' in Config and value is not None:
				value = Config['type'](value)

			if 'hook' in Config:
				ret = Config['hook'](value) if not Config.get('live') else Config['hook'](value, 'set')

				if ret is not None: # Hooks can manipulate the passed values and return them to be stored.
					value = ret

			self._set(route, value)

	def __getitem__(self, route): # #Pending: Fix: The contructed object is considered to be an unsubscriptable-object by pylint, though a __getitem__ is present.
		return self.var(route)

	def setup(self, overwrite=False):
		for route in self._Configs['']['Routes']:
			if overwrite or route not in self._Values:
				self.get(route, overwrite)

			else:
				print '%s: %s' % (route, self._Values[route])

	def get(self, route, overwrite=False):
		prefix = '  ' * route.count('/')
		Config = self._Configs.get(route)

		if not Config: # We've got a read only value.
			print '%s%s: %s' % (prefix, getLeaf(route), self._Values[route])
			return

		Routes = Config.get('Routes')

		if Routes:
			name = Config.get('name')

			if name: #  # We've got a branch
				print '\n%s%s:' % (prefix, name) # #Note" Tabs aren't used for branch indention, due the space constrains of the terminal.

				for route in Routes:
					self.get(route, overwrite)

				print ''

			# We've got a parsed value. Hence return without doing anything.

		else:
			Values = self._Values

			if route not in Values:
				self._get(route, combine(Config, {'prefix': prefix}))

			elif overwrite:
				self._get(route, combine(Config, {'default': Values[route], 'prefix': prefix})) # Have the existing value as the default.

			else:
				print '%s%s: %s' % (prefix, getLeaf(route), Values[route])

	def _get(self, route, Config):
		self._set(route, get(**Config)) # Get the input from the user and write it to the DB.

	def _set(self, route, value):
		self._Store.set({'route': route, 'value': dumps(value)}) # Set the value in the DB.
		self._Values[route] = value # Set the value in the Cache.

	def dump(self, route=''):
		Routes = self._Configs[route]['Routes']

		for route in Routes:
			Config = self._Configs.get(route)
			keyText = '%s%s' % ('  ' * route.count('/'), getLeaf(route))

			if Config and 'Routes' in Config:
				print '\n%s:' % keyText
				self.dump(route)
				print ''

			else:
				print '%s: %s' % (keyText, self._Values.get(route))

	def close(self):
		self._Store.close()

	def reopen(self):
		self._Store.reopen()

# Exports
## Functions
def getStore(**Config):
	r"""
	Returns a shared (read-only) Store.

	Config:
		filePath (str): The path to the store.
		tableName (str): The table name of the store, defaults to ecstore.
	"""
	return ReadOnlyStore(**Config)

## Config Decorators
def root(Cls=None, **Config):
	r"""
	Returns a Store for the branches and vars under the decorated class.

	Config:
		filePath: The path to the store.
		tableName (str): The table name of the store, defaults to ecstore.
		noAutoSetup (bool): Skips the auto-setup (setting up the data, when the script is inovked directly).
	"""
	if not Cls: # The decorator has some config. Hence return a wrapper to process the following class.
		return lambda Cls: root(Cls, **Config)

	Collected = collectChildren(Cls)
	Configs = {}
	Buffer = {'Configs': Configs, 'Values': {}}

	processCollected(Collected, '', Buffer)
	Configs[''] = {'Routes': Collected.keys()} # Add the root config.

	Store = ConfiguredStore(Buffer, Config)

	if not Config.get('noAutoSetup') and Cls.__module__ == '__main__': # Setup the store when the Config script is started as the main script.
		processCommand(Store)

	return Store

def branch(Obj, **Config):
	r"""Decorates the Classes that holds other branches and vars.
	"""
	State.append((getName(Obj, Config), {'Children': collectChildren(Obj), 'Config': Config}, None,))

def var(hook=None, **Config):
	r"""A decorator / function for declaring vars.
	Properties are declared using var as a function.
	Properties with hooks are defined with using this function as a decorator.

	Config is as same as ec's ArgConfig.
	"""
	if hook:
		Config['hook'] = hook
		name = getName(hook, Config)
		ret = hook

	else: # We've got a simple value or a decorator of a hook.
		name = Config.get('name', None)
		ret = lambda hook, **dummy: var(hook, **State.pop()[1]) # If the lambda is called, then it would be by the passing of the hook.

	State.append((name, Config, ret,))

	return ret

## Config Functions
def parse(filePath, format=None):
	r"""Adds parsed data to the Store, in read-only mode.

	Args:
		Check parser.parse.
	"""
	if not exists(filePath):
		raise Exception('No such file: %s' % filePath)

	Parsed = _parse(filePath, format)
	ret = lambda x: x # #Note: The lambda serves as an object marker for the assigned value, during child collection.

	State.append((None, {'Data': Parsed}, ret,))

	return ret

def value(val):
	r"""Adds a simple read-only value to the store.
	"""
	ret = lambda x: x # #Note: The lambda serves as an object marker for the assigned value, during child collection.

	State.append((None, {'value': val}, ret,))

	return ret

def data(Dict):
	r"""Adds the given dictionary into the store.
	"""
	ret = lambda x: x

	State.append((None, {'Data': Dict}, ret,))

	return ret

def store(**Config):
	r"""Returns the data from the given store, so that it could be to the root store, in read-only mode.

	Args:
		Check the call *getStore*.
	"""
	return data(getStore(**Config)[''])



r"""
initializer
===========

	Intializes the Project and gives it the Config.
"""
# State
Project = None

# Helpers
def getAttrDict(Obj, *Attrs):
	Ret = {}

	for attr in Attrs:
		if hasattr(Obj, attr):
			Ret[attr] = getattr(Obj, attr)

	return Ret

def setCWD(cwd):
	if cwd:
		from os import chdir

		chdir(cwd)

def addPaths(Paths):
	r"""Adds additional paths for module lookup.
	"""
	if Paths:
		from sys import path
		from os.path import abspath

		Paths.reverse()

		for item in Paths:
			path.insert(0, abspath(item))

def collectConfigData(Attrs):
	r"""Collects the config from various sources and builds the Config.
	"""
	from laufire.yamlex import YamlEx

	configPath = Attrs.get('configPath')

	Config = YamlEx(configPath, loglevel='ERROR') if configPath else YamlEx(loglevel='ERROR')

	if 'ConfigExtensions' in Attrs:
		Config.extend(Attrs['ConfigExtensions'])

	if 'fsRoot' in Attrs:
		from os.path import abspath
		Config['fsRoot'] = abspath(Attrs['fsRoot'])

	if 'Store' in Attrs:
		Store = Attrs['Store']
		Config.extend(Store if isinstance(Store, dict) else Store.var('')) # Note: A store could either be a dict or an ecstore.

	Config.interpolate()

	return Config

def silenceLoggers(LoggerNames):
	# #Later: Fix: Logging seems not to be silenced.

	if not LoggerNames:
		return

	import logging

	for name in LoggerNames:
		logging.getLogger(name).setLevel('ERROR') # only log really bad events

def addDefaults():
	# Load the project with default values for the essential attributes.
	from laufire.setup import Defaults

	for key, value in Defaults.iteritems():
		if not hasattr(Project, key):
			setattr(Project, key, value)

def addDevBuiltins():
	if not Project.devMode:
		return

	from laufire import dev
	import __builtin__

	for attr in [attr for attr in dir(dev) if attr[0] != 0]: # Add all the attributes of the module, dev to built-ins.
		setattr(__builtin__, attr, getattr(dev, attr))

def setSettings():
	from sys import modules

	for moduleName in ['flow', 'filesys', 'logger']:
		moduleName = 'laufire.%s' % moduleName

		if moduleName in modules:
			modules[moduleName].setup(Project)

def loadProjectSettings(setupCall):
	if Project:
		setupCall(Project)

# Init
def init():
	r"""Initializes the project.

	# #Note: The Project file has to be in the path for proper initialization.
	"""
	import Project as _Project

	global Project
	Project = _Project

	Attrs = getAttrDict(Project, 'cwd', 'Paths', 'configPath', 'fsRoot', 'ConfigExtensions', 'Store', 'LoggersToSilence')

	setCWD(Attrs.get('cwd'))

	addPaths(Attrs.get('Paths'))

	silenceLoggers(Attrs.get('LoggersToSilence')) # #Later: Think of silencing every other log, than that of the project or those that are excluded.

	Project.Config = Config = collectConfigData(Attrs)

	addDefaults()

	addDevBuiltins()

	setSettings()

	return Config

def stealCWD(scriptPath):
	from os.path import dirname

	setCWD(dirname(scriptPath))



r"""
extensions
==========
Functions to ease development.
"""
import sys

# Helpers
def getCallingModule():
	return sys.modules[sys._getframe().f_back.f_back.f_globals['__name__']] #pylint: disable=W0212

# Exports
def namespace(wrapper):
	r"""A decorator that allows functions to have private variables. This helps with simulating sub modules.
	"""
	returned = wrapper()

	if isinstance(returned, tuple): # we've got multiple items hence set the attributees of the module directly, instead of returning a decorated function

		CallingModule = getCallingModule()

		for item in returned:
			setattr(CallingModule, item.__name__, item)

	else: # copy the function signature from the wrapper to the decorated function.
		returned.__name__ = wrapper.__name__
		returned.__doc__ = wrapper.__doc__
		returned.__dict__.update(wrapper.__dict__)

		return returned

def isIterable(obj):
	return hasattr(obj, 'iteritems') or hasattr(obj, '__iter__')

def pairs(Iterable):
	r"""Provides a generator to iterate over key, value pairs of iterables.
	"""
	return Iterable.iteritems() if hasattr(Iterable, 'iteritems') else enumerate(Iterable)

def values(Iterable):
	r"""Provides a generator to iterate over key, value pairs of iterables.
	"""
	return Iterable.values() if hasattr(Iterable, 'values') else Iterable

def flatten(Iterable, recurse=False):
	r"""Collects all the values from the given iterable.

	#From: http://stackoverflow.com/questions/13490963/how-to-flatten-a-nested-dictionary-in-python-2-7x
	"""
	l = []

	for v in values(Iterable):
		if isIterable(v):
			for item in (flatten(v) if recurse else values(v)):
				l.append(item)
		else:
			l.append(v)

	return l

def combine(*Dicts):
	r"""
	Combines values from multiple dictionaries and returns a new dictionary. The values are overwritten by every following dictionary.
	"""
	Ret = {}

	for Dict in Dicts:
		Ret.update(Dict)

	return Ret

def merge(*Dicts):
	r"""
	Recursively merges the passed dictionaries.
	"""
	Ret = {}

	for Dict in Dicts:
		for key, value in Dict.iteritems():
			if key in Ret:
				tgtValue = Ret[key]

				if hasattr(tgtValue, 'iteritems') and hasattr(value, 'iteritems'):
					value = merge(tgtValue, value)

				Ret[key] = value

			else:
				Ret[key] = value

	return Ret

def select(Dict, Keys):
	r"""
	Returns a sub-dictionary of the given Dict with the given keys.
	"""
	return {k: Dict[k] for k in Keys}

def unpack(Dict, *Keys):
	r"""
	Unpacks the given keys of the given dictionary and assign them to the given variables.

	#Ex: a, b, c = unpack(Dict, 'a', 'b', 'c')
	"""
	return tuple(Dict[k] for k in Keys)

def walk(Iterable, RouteParts=None):
	if RouteParts is None:
		RouteParts = [None]

	for key, val in pairs(Iterable):
		RouteParts[-1 if RouteParts else 0] = key

		if isIterable(val):
			for val, RouteParts, Iterable in walk(val, [key]):
				yield val, [key] + RouteParts, Iterable

		else:
			yield val, RouteParts, Iterable

def unnest(Dict, separator='/', Target=None):
	r"""Gets a routed dictionary from a nested one.
	"""
	if Target is None:
		Target = {}

	for val, RouteParts, dummy in walk(Dict):
		Target[separator.join(RouteParts)] = val

	return Target

def resolveRoute(Dict, route, separator='/'):
	r"""Resolves a route from a nested dict.

	Args:
		Dict (dict): The dict to look up for the route.
		route (str / list): The route to resolve.
		separator (str): Defaults to '/'.
	"""
	if not route:
		return Dict

	KeyParts = route.split(separator) if hasattr(route, 'split') else route

	Resolved = Dict

	for item in KeyParts:
		Resolved = Resolved[item]

	return Resolved

class Lazy:
	r"""A class to implement lazy initialization. The class passed during the initialization will be initialized on first access.
	"""
	def __init__(self, Underlying, *Args, **KWArgs):
		self._data = Underlying, Args, KWArgs
		self._obj = None

	def __getattr__(self, name):
		if self._obj:
			return getattr(self._obj, name)

		Underlying, Args, KWArgs = self._data
		self._obj = Underlying(*Args, **KWArgs) #pylint: disable=W0201
		delattr(self, '_data')

		return getattr(self, name)



r"""
gitcli
======

A module to help with git related tasks. It depends upon the availability of the 'git' command through the command line. Existing solutions aren't chosen to save the time to learn.

Notes:

	* **cwd**, **HEAD** and **the current branch** are assumed to be the defaults.

# #Pending: Use a class, so that commands do not have to depend upon the CWD.
"""
from subprocess import Popen, PIPE
from shlex import split
from os.path import abspath

# Helpers
def call(command, **kwargs): # from gitapi.py
	r"""Executes a shell command and returns its outputs.
	"""
	proc = Popen(split(command), stdout=PIPE, stderr=PIPE, **kwargs)
	out, err = [x.decode("utf-8") for x in proc.communicate()]

	return {'out': out, 'err': err, 'code': proc.returncode}

def getStdout(commandStr, **kwargs):
	r"""Returns the stripped stdout of a successful command.
	"""
	Result = call('git %s' % commandStr, **kwargs)

	if Result['code'] != 0:
		raise Exception(Result['err'])

	return Result['out'].strip()

# Exports
## Low level calls
def command(commandStr, cwd='.'):
	r"""Executes a git command and returns its outputs.
	"""
	return call('git %s' % commandStr, cwd=cwd)

def lines(commandStr, cwd='.'):
	r"""Executes a git command and returns its stdout lines as a list.
	"""
	Result = call('git %s' % commandStr, cwd=cwd)

	if Result['code']:
		errStr = Result['err'].strip()
		raise(Exception(errStr[:errStr.find('\n')]))

	out = Result['out'].strip()

	return out.split('\n') if out else []

## Info
def isClean(dir='.'):
	r"""Ensures that the given dir doesn't have any uncommited changes.
	"""
	cwd = getRepoPath(dir)
	dir = abspath(dir)

	return getStdout('status "%s" --porcelain' % dir, cwd=cwd) == ''

def getCommitMessage(dir='.'):
	r"""Returns the commit message of the last commit of the current branch.
	"""
	return getStdout('log -1 --pretty=%B', cwd=dir).strip()

def getCommitID(dir='.', short=0):
	r"""Returns the commit id of the last commit of the current branch.
	"""
	return getStdout('rev-parse %s HEAD' % (('--short=%s' % short) if short else ''), cwd=dir).strip()

def getCurrentBranch(dir='.'):
	r"""Returns the name of the current branch.
	"""
	return getStdout('rev-parse --abbrev-ref HEAD', cwd=dir).strip()

def getFileList(dir='.'):
	r"""Returns the list of gitted files under the given dir.
	"""
	return getStdout('ls-files', cwd=dir).split('\n')

def getRepoPath(dir='.'):
	return getStdout('rev-parse --show-toplevel', cwd=dir).strip()

def isExportable(dir='.'):
	r"""Ensures that the current state of the repo is clean and of the master branch.
	"""
	return isClean(getRepoPath(dir)) and getCurrentBranch(dir) == 'master'

# Tagging
def getTags(pattern=None, dir='.'):
	r"""Returns the list of tags from the current branch.
	"""
	return lines('tag%s' % ('' if not pattern else ' -l "%s"' % pattern), cwd=dir)

def addTag(tagName, tgtTreeish=None, dir='.'):
	getStdout('tag "%s"%s' % (tagName, '' if not tgtTreeish else (' "%s"' % tgtTreeish)), cwd=dir)

def deleteTag(tagName, dir='.'):
	getStdout('tag -d "%s"' % tagName, cwd=dir)

## Utilities
def archive(archiveTarget, pathToArchive='.', treeish='HEAD'):
	r"""Archives the pathToArchive of the given treeish, to the archiveTarget.
	"""
	assert(command('archive --format=zip --output "%s" %s' % (archiveTarget, treeish), cwd=pathToArchive)['code'] == 0)



r"""
YamlEx
======

A wrapper around HiYaPyCo that supports:

* Easier file handling.
* Runtime variables.

#Later: Allow external template sources. ie: The source data won't be in the rendered data. This passibly could be a KWArg named, env.
"""
from hiyapyco import HiYaPyCo, dump, odyldo
from jinja2 import Template

from .extensions import merge

# #Later: Make YamlEx to inherit 'dict', so that it could support all the functionalities of a dict.
class YamlEx:
	r"""The wrapper class.

	Args:
		*FilePaths: A list of YAML files to load.
	"""
	def __init__(self, *FilePaths, **KWArgs):
		if FilePaths and FilePaths[0] is None: # None is supplied for the file path.
			FilePaths = None

		self.HiYaPyCo = H = HiYaPyCo(*FilePaths, **KWArgs) if FilePaths else HiYaPyCo('---\n{}', **KWArgs) # return an empty object when no file is mentioned.
		self.Data = H._data #pylint: disable=W0212

	def __get_item__(self, key):
		return self.Data[key]

	def setData(self, Data):
		r"""Explicitly sets self.Data.

		Note:
			An explicit set method is used instead of getters and setters, as the way to implement them in this context coudn't be found (@property + new style classes, meddles with __getitem__).
		"""
		self.Data = Data
		self.HiYaPyCo._data = Data #pylint: disable=W0212

		return self

	def extend(self, DataToMerge):
		r"""
		Extends the existing data by merging it with the given data.
		"""
		from laufire.extensions import merge

		Data = merge(self.Data, DataToMerge)
		self.Data = Data
		self.HiYaPyCo._data = Data #pylint: disable=W0212

		return self

	def interpolate(self):
		r"""Does the interpolation.
		"""
		self.HiYaPyCo._interpolate(self.Data) #pylint: disable=W0212
		return self

	def load(self, filePath, interpolate=True):
		r"""Loads data from the given YAML file.

		Args:
			filePath (str): A path to a valid YAML file.
		"""
		self.setData(merge(self.Data, odyldo.safe_load(open(filePath, 'r'))))

		if interpolate:
			self.interpolate()

		return self

	def dump(self, defaultFlowStyle=True):
		r"""Returns the YAML dump.
		"""
		return dump(self.Data, defaultFlowStyle)

	def render(self, tmplStr):
		r"""Renders the given template.
		"""
		return Template(tmplStr).render(self.Data)

	# Allow access to the underlying dictionary attrs, this also will allow access to the keys of self.Data.
	def __getattr__(self, attr):
		return getattr(self.Data, attr)

def overlayHiYaPyCo():
	import hiyapyco
	from jinja2 import Environment, Undefined
	from jinja2.utils import missing
	from jinja2._compat import implements_to_string

	# Allow the undefined vars to be as is
	@implements_to_string
	class UntouchedUndefined(Undefined):
		r"""Skips the undefined variables during interpolation, to allow runtime variable assignments.
		"""
		__slots__ = ()

		def __str__(self):
			if self._undefined_obj is missing:
				return u'{{ %s }}' % self._undefined_name

			return self

		__iter__ = __len__ = __nonzero__ = __eq__ = __ne__ = __bool__ = __hash__ = __str__

	hiyapyco.jinja2env = Environment(undefined=UntouchedUndefined)

	# Hook into HiYaPyCo._interpolatestr so that string could be interpolated recursively.

	HiYaPyCo._interpolatestr_orig = HiYaPyCo._interpolatestr #pylint: disable=W0212

	def _interpolatestrHook(self, s):
		rendered = self._interpolatestr_orig(s) #pylint: disable=W0212

		while s != rendered:
			s = rendered
			rendered = self._interpolatestr_orig(s) #pylint: disable=W0212

		return s

	HiYaPyCo._interpolatestr = _interpolatestrHook #pylint: disable=W0212

if not hasattr(HiYaPyCo, '_interpolatestr_orig'): # Skip overlaying, if the module is parsed more than one time (through import collision etc).
	overlayHiYaPyCo()



r"""
A module to help with encrypting and decrypting the strings.

#From: The module is improved over the script ccavutil.py from CCAvenue's integration kit for python.
"""
import md5
import pyaes

# Data
iv = '\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'

# Helpers
def pad(data):
	length = 16 - (len(data) % 16)
	data += chr(length) * length
	return data

def unpad(data):
	return data[:-ord(data[len(data)-1])]

def getAES(key):
	digest = md5.new()
	digest.update(key)

	return pyaes.AESModeOfOperationCBC(digest.digest(), iv=iv)

def spliceString(string, length=1):
	for block in [string[i:i+length] for i in range(0, len(string), length)]:
		yield block

# Exports
def encrypt(plainText, key):
	plainText = pad(plainText)
	aes = getAES(key)
	encryptedText = ''

	for block in spliceString(plainText, 16):
		encryptedText += aes.encrypt(block)

	return encryptedText.encode('hex')

def decrypt(cipherText, key):
	encryptedText = cipherText.decode('hex')
	aes = getAES(key)
	decryptedText = ''

	for block in spliceString(encryptedText, 16):
		decryptedText += aes.decrypt(block)

	return unpad(decryptedText)



r"""
flow
----

A module to control the flow of the application.
"""
from time import sleep

from laufire.logger import debug, log

defaultDelay = 0
tickTime = 0.1

def waitFor(func, maxWait=None, message=None):
	r"""Waits for the given function to return a truthy or until a set time.
	"""
	waited = 0
	maxWait = maxWait or defaultDelay

	if message:
		log(message)

	while True:

		ret = func()

		if ret:
			return ret

		waited += tickTime

		if waited > maxWait:
			raise Exception('Maximum wait time exceded.' if not message else 'Failed waiting for: %s' % message)

		if message:
			debug(message)

		debug('Waited: %d, MaxWait: %d.' % (waited, maxWait))
		sleep(tickTime)

def forgive(func):
	try:
		func()

	except Exception as e: #pylint: disable=W0703
		debug(e)
		return e

def rob(func):
	r"""Tries to get the result from a given function, and returns none if there were an error.
	"""
	try:
		return func()

	except: #pylint: disable=W0702
		return

def retry(func, repeat=3, delay=tickTime * 2):
	r"""
	Calls the given function till it returns a value.

	Set repeat to -1, to try untill success.
	"""
	while repeat:
		result = func()

		if result is None and delay and repeat != 1:
			sleep(delay)

		else:
			return result

		repeat -= 1

# Init
def setup(Project):
	delay = Project.delay / 1.0 # Get a float value.

	global defaultDelay, tickTime

	defaultDelay = delay
	tickTime = defaultDelay / 10

from laufire.initializer import loadProjectSettings #pylint: disable=C0413
loadProjectSettings(setup)



r"""
A module to help with logging.

The reason for a custom module is due to different preferences in the output levels and style.
"""

import logging

from colorama import Fore, Style, init as colorama_init

# Delegates
Logger = logging.getLogger('<unnamed>')

# Exports
__all__ = ['log', 'logError', 'debug']

# Data
BRED = '%s%s' % (Style.BRIGHT, Fore.RED) #pylint: disable=E1101

# State
Supressed = []

def log(message, color=None):
	r"""Facilitates colored logging.
	"""
	Logger.info('%s%s' % (getattr(Fore, color, 'WHITE'), message) if color else message)

def logError(message):
	r"""Logs an error.
	"""
	Logger.error('%s%s\n' % (BRED, message)) #pylint: disable=W1201

def dump(message):
	if Logger.level > 1:
		Supressed.append(message)

	else:
		Logger.debug(message)

def debug(message):
	if Logger.level > 10:
		Supressed.append(message) #Pending: Add a max-size setting for the Buffer.

	Logger.debug(message)

def setLevel(lvl):
	r"""Sets the level of the Logger.

		Args:
			lvl	 (int): Could be one of the following integers (1, 2, 3, 4, 5). The greater the number lesser the logs.
	"""
	Logger.setLevel((lvl * 10) or 1) # #Note: 0 silences the logger.

# Init
def setup(Project):
	Logger.name = Project.name
	setLevel(Project.logLevel)

def init():
	colorama_init(autoreset=True)

	from laufire.initializer import loadProjectSettings
	loadProjectSettings(setup)

	Logger.addHandler(logging.StreamHandler())
	Logger.setLevel(logging.INFO)

init()



r"""
SS
==

	A simple shell, that helps to interact with the given function.
"""
def interact(func, prompt='>'):
	r"""
	Interact with the given function. The interaction can be closed by typing the EOF character.
	"""
	while True:
		try:
			func(raw_input(prompt))

		except EOFError:
			break

		except Exception as e: #pylint: disable=broad-except
			print e



r"""
A module with constructors for mockable objects and their mocks.
"""

# State
Gateway = None

def getGateway(Config, shouldMock, cached=True):
	r"""
	Returns a Gateway according to the passed params.

	Args:
		Config (dict): The Config as expected (should have SSH, Gateway and Mock configs).
		shouldMock (bool): Returns the mocker when set to True.
		cached (bool): Shares a cached Gateway instead of creating a new one with every call.
	"""
	if cached:
		global Gateway

		if Gateway:
			return Gateway

	if shouldMock:
		from laufire.mock.ssh import SSHBridgeMocker
		Gateway = SSHBridgeMocker(Config)

	else:
		from laufire.ssh import SSHBridge
		Gateway = SSHBridge(Config)

	return Gateway
